@@ -52,6 +52,7 @@
 #include <linux/rtnetlink.h>
 #include <unistd.h>
 #include <sys/user.h>
+#include <sys/eventfd.h>
 #include "cryptouser.h"
 
 #include "kcapi.h"
@@ -60,10 +61,10 @@
 		      * require consumer to be updated (as long as this number
 		      * is zero, the API is not considered stable and can
 		      * change without a bump of the major version) */
-#define MINVERSION 7 /* API compatible, ABI may change, functional
+#define MINVERSION 8 /* API compatible, ABI may change, functional
 		      * enhancements only, consumer can be left unchanged if
 		      * enhancements are not considered */
-#define PATCHLEVEL 2 /* API / ABI compatible, no functional changes, no
+#define PATCHLEVEL 0 /* API / ABI compatible, no functional changes, no
 		      * enhancements, bug fixes only */
 
 /* remove once in if_alg.h */
@@ -82,8 +83,6 @@
 #define SOL_ALG 279
 #endif
 
-#define MAXPIPELEN (16 * PAGE_SIZE)
-
 /************************************************************
  * Internal logic
  ************************************************************/
@@ -94,7 +93,6 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 {
 	ssize_t ret = -EINVAL;
 	char *buffer = NULL;
-	volatile void *_buffer = NULL;
 	int errsv = 0;
 
 	/* plaintext / ciphertext data */
@@ -162,11 +160,7 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 	ret = sendmsg(handle->opfd, &msg, flags);
 	errsv = errno;
 
-	memset(buffer, 0, bufferlen);
-	/* magic to convince GCC to memset the buffer */
-	_buffer = memchr(buffer, 1, bufferlen);
-	if (_buffer)
-		_buffer = '\0';
+	kcapi_memset_secure(buffer, 0, bufferlen);
 	free(buffer);
 	return (ret >= 0) ? ret : -errsv;
 }
@@ -190,30 +184,55 @@ static inline ssize_t _kcapi_common_send_data(struct kcapi_handle *handle,
 	return (ret >= 0) ? ret : -errno;
 }
 
-static inline ssize_t _kcapi_common_vmsplice_data(struct kcapi_handle *handle,
-						  struct iovec *iov,
-						  size_t iovlen,
-						  size_t inlen,
-						  unsigned int flags)
+static inline ssize_t _kcapi_common_vmsplice_iov(struct kcapi_handle *handle,
+						 struct iovec *iov,
+						 unsigned long iovlen)
 {
 	ssize_t ret = 0;
+	size_t inlen = 0;
+	unsigned long i;
 
-	if (inlen > MAXPIPELEN) {
-		fprintf(stderr, "Splice operation: input data exceeds maximum allowed size of %lu\n",
-			MAXPIPELEN);
-		return -E2BIG;
-	}
+	for (i = 0; i < iovlen; i++)
+		inlen += iov[i].iov_len;
 
-	ret = vmsplice(handle->pipes[1], iov, iovlen, SPLICE_F_GIFT|flags);
+	ret = vmsplice(handle->pipes[1], iov, iovlen, SPLICE_F_GIFT);
 	if (0 > ret)
 		return ret;
 	if ((size_t)ret != inlen)
 		fprintf(stderr, "vmsplice: not all data received by kernel (data recieved: %ld -- data sent: %lu)\n",
 			(long)ret, (unsigned long)inlen);
-	ret = splice(handle->pipes[0], NULL, handle->opfd, NULL, ret, flags);
+	ret = splice(handle->pipes[0], NULL, handle->opfd, NULL, ret, 0);
 	return (ret >= 0) ? ret : -errno;
 }
 
+static inline ssize_t _kcapi_common_vmsplice_chunk(struct kcapi_handle *handle,
+						   const unsigned char *in,
+						   size_t inlen)
+{
+	struct iovec iov;
+	size_t processed = 0;
+
+	while (inlen) {
+		ssize_t ret = 0;
+
+		iov.iov_base = (void*)(uintptr_t)(in + processed);
+		iov.iov_len = inlen;
+		ret = vmsplice(handle->pipes[1], &iov, 1,
+			       SPLICE_F_GIFT|SPLICE_F_MORE);
+		if (0 > ret)
+			return ret;
+		ret = splice(handle->pipes[0], NULL, handle->opfd, NULL, ret,
+			     SPLICE_F_MORE);
+		if (0 > ret)
+			return ret;
+
+		processed += ret;
+		inlen -= ret;
+	}
+
+	return processed;
+}
+
 static inline ssize_t _kcapi_common_recv_data(struct kcapi_handle *handle,
 					      struct iovec *iov, size_t iovlen)
 {
@@ -451,8 +470,32 @@ static int _kcapi_common_getinfo(struct kcapi_handle *handle,
 	return 0;
 }
 
-static int _kcapi_handle_init(struct kcapi_handle *handle,
-			      const char *type, const char *ciphername)
+/* Initialize AIO -- on error, AIO is simply not used, but lib lives on */
+static void _kcapi_aio_init(struct kcapi_handle *handle)
+{
+	if (handle->aio.skcipher_aio_disable)
+		return;
+
+	handle->aio.efd = eventfd(0, EFD_CLOEXEC);
+	if (handle->aio.efd < 0)
+		goto err;
+
+	if (io_setup(KCAPI_AIO_CONCURRENT, &handle->aio.aio_ctx) < 0) {
+		perror("io_setup error");
+		goto err;
+	}
+	return;
+
+err:
+	handle->aio.skcipher_aio_disable = 1;
+	if (handle->aio.efd != -1)
+		close(handle->aio.efd);
+	handle->aio.efd = -1;
+	return;
+}
+
+static int _kcapi_handle_init(struct kcapi_handle *handle, const char *type,
+			      const char *ciphername, unsigned int flags)
 {
 	struct sockaddr_alg sa;
 	int ret;
@@ -504,10 +547,25 @@ static int _kcapi_handle_init(struct kcapi_handle *handle,
 		close(handle->pipes[0]);
 		close(handle->pipes[1]);
 	}
+
+	if (flags & KCAPI_INIT_AIO)
+		_kcapi_aio_init(handle);
+	else
+		handle->aio.skcipher_aio_disable = 1;
+
 	return (errsv) ? -errsv : ret;
 }
 
-static inline int _kcapi_handle_destroy(struct kcapi_handle *handle)
+static inline void _kcapi_aio_destroy(struct kcapi_handle *handle)
+{
+	if (handle->aio.skcipher_aio_disable)
+		return;
+	if (handle->aio.efd != -1)
+		close(handle->aio.efd);
+	io_destroy(handle->aio.aio_ctx);
+}
+
+static inline void _kcapi_handle_destroy(struct kcapi_handle *handle)
 {
 	if (handle->tfmfd != -1)
 		close(handle->tfmfd);
@@ -517,24 +575,16 @@ static inline int _kcapi_handle_destroy(struct kcapi_handle *handle)
 		close(handle->pipes[0]);
 	if (handle->pipes[1] != -1)
 		close(handle->pipes[1]);
+	_kcapi_aio_destroy(handle);
 	memset(handle, 0, sizeof(struct kcapi_handle));
-	return 0;
 }
 
-static inline size_t _kcapi_aead_encrypt_outlen(struct kcapi_handle *handle,
-						size_t inlen, size_t taglen)
+/*********** Generic Helper functions *************************/
+
+void kcapi_memset_secure(void *s, int c, size_t n)
 {
-	int bs = handle->info.blocksize;
-
-	return ((inlen + bs - 1) / bs * bs + taglen);
-}
-
-static inline size_t _kcapi_aead_decrypt_outlen(struct kcapi_handle *handle,
-						size_t inlen)
-{
-	int bs = handle->info.blocksize;
-
-	return ((inlen + bs - 1) / bs * bs);
+	memset(s, c, n);
+	__asm__ __volatile__("" : : "r" (s) : "memory");
 }
 
 void kcapi_versionstring(char *buf, size_t buflen)
@@ -566,8 +616,9 @@ int kcapi_pad_iv(struct kcapi_handle *handle,
 	ret = posix_memalign((void *)&niv, 16, nivlen);
 	if (ret)
 		return -ret;
-	memset(niv, 0, nivlen);
 	memcpy(niv, iv, copylen);
+	if (nivlen > copylen)
+		memset(niv + copylen, 0, nivlen - copylen);
 
 	*newiv = niv;
 	*newivlen = nivlen;
@@ -575,14 +626,15 @@ int kcapi_pad_iv(struct kcapi_handle *handle,
 	return 0;
 }
 
-int kcapi_cipher_init(struct kcapi_handle *handle, const char *ciphername)
+int kcapi_cipher_init(struct kcapi_handle *handle, const char *ciphername,
+		      unsigned int flags)
 {
-	return _kcapi_handle_init(handle, "skcipher", ciphername);
+	return _kcapi_handle_init(handle, "skcipher", ciphername, flags);
 }
 
-int kcapi_cipher_destroy(struct kcapi_handle *handle)
+void kcapi_cipher_destroy(struct kcapi_handle *handle)
 {
-	return _kcapi_handle_destroy(handle);
+	_kcapi_handle_destroy(handle);
 }
 
 int kcapi_cipher_setkey(struct kcapi_handle *handle,
@@ -591,14 +643,99 @@ int kcapi_cipher_setkey(struct kcapi_handle *handle,
 	return _kcapi_common_setkey(handle, key, keylen);
 }
 
+static inline ssize_t _kcapi_aio_read(struct kcapi_handle *handle,
+				     unsigned char *out, size_t outlen)
+{
+	(void)handle;
+	(void)out;
+	(void)outlen;
+	fprintf(stderr, "AIO support not complete\n");
+	return -EOPNOTSUPP;
+#if 0
+	struct iocb *cb = NULL;
+	int r = 0;
+
+	cb = &cbt[handle->aio.curr_cio];
+        while (cb->aio_fildes)
+		poll_data(10);
+
+        memset(cb, 0, sizeof(*cb));
+        cb->aio_fildes = handle->opfd;
+        cb->aio_lio_opcode = IOCB_CMD_PREAD;
+        cb->aio_buf = out;
+        cb->aio_offset = 0;
+        cb->aio_data = i;
+        cb->aio_nbytes = outlen;
+        cb->aio_flags = IOCB_FLAG_RESFD;
+        cb->aio_resfd = handle->aio.efd;
+        r = io_read(handle->aio.aio_ctx, 1, &cb);
+        if (r != 1) {
+		if (r < 0) {
+			fprintf(stderr, "io_read Error: %d\n", errno);
+			return -EFAULT;
+		} else {
+			fprintf(stderr, "Could not sumbit AIO read\n");
+			return -EIO;
+		}
+	}
+
+        if (handle->aio.curr_cio) >= KCAPI_AIO_CONCURRENT / 2)
+		poll_data(1);
+
+	return 0;
+#endif
+}
+
+static ssize_t _kcapi_cipher_crypt(struct kcapi_handle *handle,
+				   const unsigned char *in, size_t inlen,
+				   unsigned char *out, size_t outlen,
+				   int access, int enc)
+{
+	struct iovec iov;
+	ssize_t ret = 0;
+
+
+	iov.iov_base = (void*)(uintptr_t)in;
+	/*
+	 * Using two syscalls with memcpy is faster than four syscalls
+	 * without memcpy below the given threshold.
+	 */
+	if ((access == KCAPI_ACCESS_HEURISTIC && inlen <= (1<<13)) ||
+	    access == KCAPI_ACCESS_SENDMSG) {
+		iov.iov_len = inlen;
+		ret = _kcapi_common_send_meta(handle, &iov, 1, enc, 0);
+		if (0 > ret)
+			return ret;
+		iov.iov_base = (void*)(uintptr_t)out;
+		iov.iov_len = outlen;
+	} else {
+		ret = _kcapi_common_send_meta(handle, NULL, 0, enc, 0);
+		if (0 > ret)
+			return ret;
+		ret = _kcapi_common_vmsplice_chunk(handle, in, inlen);
+		if (0 > ret)
+			return ret;
+	}
+
+	if (handle->aio.skcipher_aio_disable) {
+		return _kcapi_common_read_data(handle, out, outlen);
+	} else {
+		ret = _kcapi_aio_read(handle, out, outlen);
+		if (ret == -EIO) {
+			/* AIO not available for specific interface */
+			handle->aio.skcipher_aio_disable = 1;
+			_kcapi_aio_destroy(handle);
+			return _kcapi_common_recv_data(handle, &iov, 1);
+		}
+		return ret;
+	}
+}
+
 ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 			     const unsigned char *in, size_t inlen,
 			     const unsigned char *iv,
 			     unsigned char *out, size_t outlen, int access)
 {
-	struct iovec iov;
-	ssize_t ret = 0;
-	size_t processed = 0;
 	unsigned int bs = handle->info.blocksize;
 
 	if (!in || !inlen || !out || !outlen) {
@@ -616,41 +753,8 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 	}
 
 	handle->cipher.iv = iv;
-
-	iov.iov_base = (void*)(uintptr_t)in;
-	/*
-	 * Using two syscalls with memcpy is faster than four syscalls
-	 * without memcpy below the given threshold.
-	 */
-	if ((access == KCAPI_ACCESS_HEURISTIC && inlen <= (1<<13)) ||
-	    access == KCAPI_ACCESS_SENDMSG) {
-		iov.iov_len = inlen;
-		ret = _kcapi_common_send_meta(handle, &iov, 1, ALG_OP_ENCRYPT,
-					      0);
-		if (0 > ret)
-			return ret;
-		iov.iov_base = (void*)(uintptr_t)out;
-		iov.iov_len = outlen;
-		return _kcapi_common_recv_data(handle, &iov, 1);
-	}
-
-	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT, 0);
-	if (0 > ret)
-		return ret;
-	while (inlen) {
-		size_t datalen = (inlen > MAXPIPELEN) ? MAXPIPELEN : inlen;
-
-		iov.iov_len = datalen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov, 1, datalen,
-						  SPLICE_F_MORE);
-		if (0 > ret)
-			return ret;
-		processed += ret;
-		iov.iov_base = (void*)(uintptr_t)(in + processed);
-		inlen -= ret;
-	}
-
-	return _kcapi_common_read_data(handle, out, outlen);
+	return _kcapi_cipher_crypt(handle, in, inlen, out, outlen, access,
+				   ALG_OP_ENCRYPT);
 }
 
 ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
@@ -658,10 +762,6 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 			     const unsigned char *iv,
 			     unsigned char *out, size_t outlen, int access)
 {
-	struct iovec iov;
-	ssize_t ret = 0;
-	size_t processed = 0;
-
 	if (!in || !inlen || !out || !outlen) {
 		fprintf(stderr,
 			"Symmetric Decryption: Empty plaintext or ciphertext buffer provided\n");
@@ -684,41 +784,8 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 	}
 
 	handle->cipher.iv = iv;
-
-	iov.iov_base = (void*)(uintptr_t)in;
-
-	/*
-	 * Using two syscalls with memcpy is faster than four syscalls
-	 * without memcpy below the given threshold.
-	 */
-	if ((access == KCAPI_ACCESS_HEURISTIC && inlen <= (1<<13)) ||
-	    access == KCAPI_ACCESS_SENDMSG) {
-		iov.iov_len = inlen;
-		ret = _kcapi_common_send_meta(handle, &iov, 1, ALG_OP_DECRYPT,
-					      0);
-		iov.iov_base = (void*)(uintptr_t)out;
-		iov.iov_len = outlen;
-		return _kcapi_common_recv_data(handle, &iov, 1);
-	}
-
-	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT, 0);
-	if (0 > ret)
-		return ret;
-
-	while (inlen) {
-		size_t datalen = (inlen > MAXPIPELEN) ? MAXPIPELEN : inlen;
-
-		iov.iov_len = datalen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov, 1, datalen,
-						  SPLICE_F_MORE);
-		if (0 > ret)
-			return ret;
-		processed += ret;
-		iov.iov_base = (void*)(uintptr_t)(in + processed);
-		inlen -= ret;
-	}
-
-	return _kcapi_common_read_data(handle, out, outlen);
+	return _kcapi_cipher_crypt(handle, in, inlen, out, outlen, access,
+				   ALG_OP_DECRYPT);
 }
 
 ssize_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
@@ -766,14 +833,15 @@ unsigned int kcapi_cipher_blocksize(struct kcapi_handle *handle)
 	return handle->info.blocksize;
 }
 
-int kcapi_aead_init(struct kcapi_handle *handle, const char *ciphername)
+int kcapi_aead_init(struct kcapi_handle *handle, const char *ciphername,
+		    unsigned int flags)
 {
-	return _kcapi_handle_init(handle, "aead", ciphername);
+	return _kcapi_handle_init(handle, "aead", ciphername, flags);
 }
 
-int kcapi_aead_destroy(struct kcapi_handle *handle)
+void kcapi_aead_destroy(struct kcapi_handle *handle)
 {
-	return _kcapi_handle_destroy(handle);
+	_kcapi_handle_destroy(handle);
 }
 
 int kcapi_aead_setkey(struct kcapi_handle *handle,
@@ -798,77 +866,92 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, size_t assoclen)
 	handle->aead.assoclen = assoclen;
 }
 
-ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
-			   const unsigned char *in, size_t inlen,
-			   const unsigned char *iv,
-			   const unsigned char *assoc, unsigned char *out,
-			   size_t outlen, int access)
+void kcapi_aead_getdata(struct kcapi_handle *handle,
+			unsigned char *encdata, size_t encdatalen,
+			unsigned char **aad, size_t *aadlen,
+			unsigned char **data, size_t *datalen,
+			unsigned char **tag, size_t *taglen)
 {
-	struct iovec iov[2];
+	if (encdatalen <  handle->aead.taglen + handle->aead.assoclen) {
+		fprintf(stderr, "Result of encryption operation (%lu) is smaller than tag and AAD length (%lu)\n",
+			(unsigned long)encdatalen,
+			(unsigned long)handle->aead.taglen +
+			(unsigned long)handle->aead.assoclen);
+		return;
+	}
+	if (aad) {
+		*aad = encdata;
+		*aadlen = handle->aead.assoclen;
+	}
+	if (data) {
+		*data = encdata + handle->aead.assoclen;
+		*datalen = encdatalen - handle->aead.assoclen -
+			   handle->aead.taglen;
+	}
+	if (tag) {
+		*tag = encdata + encdatalen - handle->aead.taglen;
+		*taglen = handle->aead.taglen;
+	}
+}
+
+ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
+			   unsigned char *in, size_t inlen,
+			   const unsigned char *iv,
+			   unsigned char *out, size_t outlen, int access)
+{
+	struct iovec iov[3];
 	ssize_t ret = 0;
+	int i = 0;
+	unsigned char *assoc = NULL;
+	size_t assoclen = 0;
+	unsigned char *data = NULL;
+	size_t datalen = 0;
+	unsigned char *tag = NULL;
+	size_t taglen = 0;
 
 	/* require properly sized output data size */
-	if (outlen < _kcapi_aead_encrypt_outlen(handle, inlen,
-						handle->aead.taglen) ) {
+	if (outlen < inlen) {
 		fprintf(stderr,
-			"AEAD Encryption: Ciphertext buffer (%lu) is not plaintext buffer (%lu) rounded up to multiple of block size %u plus tag length %lu\n",
-			(unsigned long)outlen, (unsigned long)inlen,
-			handle->info.blocksize,
-			(unsigned long)handle->aead.taglen);
+			"AEAD Encryption: Ciphertext buffer (%lu) is smaller than plaintext buffer (%lu)\n",
+			(unsigned long)outlen, (unsigned long)inlen);
 		return -EINVAL;
 	}
 
 	handle->cipher.iv = iv;
 
+	kcapi_aead_getdata(handle, in, inlen,
+			   &assoc, &assoclen, &data, &datalen, &tag, &taglen);
+	if (assoclen) {
+		iov[i].iov_base = (void *)assoc;
+		iov[i].iov_len = assoclen;
+		i++;
+	}
+	if (datalen) {
+		iov[i].iov_base = (void *)data;
+		iov[i].iov_len = datalen;
+		i++;
+	}
+	if (taglen) {
+		iov[i].iov_base = (void *)tag;
+		iov[i].iov_len = taglen;
+		i++;
+	}
+
 	if (access == KCAPI_ACCESS_HEURISTIC ||
 	    access == KCAPI_ACCESS_SENDMSG) {
-		if (assoc && handle->aead.assoclen) {
-			iov[0].iov_base = (void*)(uintptr_t)assoc;
-			iov[0].iov_len = handle->aead.assoclen;
-			iov[1].iov_base = (void*)(uintptr_t)in;
-			iov[1].iov_len = inlen;
-			ret = _kcapi_common_send_meta(handle, &iov[0], 2,
-						ALG_OP_ENCRYPT, 0);
-		} else {
-			iov[0].iov_base = (void*)(uintptr_t)in;
-			iov[0].iov_len = inlen;
-			ret = _kcapi_common_send_meta(handle, &iov[0], 1,
-						ALG_OP_ENCRYPT, 0);
-		}
+		ret = _kcapi_common_send_meta(handle, &iov[0], i,
+					      ALG_OP_ENCRYPT, 0);
 		if (0 > ret)
 			return ret;
-		ret = _kcapi_common_read_data(handle, out, outlen);
-		if (ret < 0)
-			return ret;
-		if ((ret < (ssize_t)handle->aead.taglen))
-			return -E2BIG;
-		return ret;
-	}
-
-	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT,
+	} else {
+		ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT,
 			(handle->aead.assoclen || inlen) ? MSG_MORE : 0);
-	if (0 > ret)
-		return ret;
-	if (assoc && handle->aead.assoclen) {
-		size_t len = 0;
-
-		iov[0].iov_base = (void*)(uintptr_t)assoc;
-		iov[0].iov_len = handle->aead.assoclen;
-		len = handle->aead.assoclen;
-		iov[1].iov_base = (void*)(uintptr_t)in;
-		iov[1].iov_len = inlen;
-		len += inlen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2, len, 0);
-	} else if (inlen) {
-		size_t len = 0;
-
-		iov[0].iov_base = (void*)(uintptr_t)in;
-		iov[0].iov_len = inlen;
-		len = inlen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 1, len, 0);
+		if (0 > ret)
+			return ret;
+		ret = _kcapi_common_vmsplice_iov(handle, &iov[0], i);
+		if (0 > ret)
+			return ret;
 	}
-	if (0 > ret)
-		return ret;
 
 	ret = _kcapi_common_read_data(handle, out, outlen);
 	if (ret < 0)
@@ -879,111 +962,66 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 	return ret;
 }
 
-void kcapi_aead_getdata(struct kcapi_handle *handle,
-			unsigned char *encdata, size_t encdatalen,
-			unsigned char **data, size_t *datalen,
-			unsigned char **tag, size_t *taglen)
-{
-	if (encdatalen <  handle->aead.taglen) {
-		fprintf(stderr, "Result of encryption operation (%lu) is smaller than tag length (%lu)\n",
-			(unsigned long)encdatalen,
-			(unsigned long)handle->aead.taglen);
-		return;
-	}
-	if (data) {
-		*data = encdata;
-		*datalen = encdatalen - handle->aead.taglen;
-	}
-	if (tag) {
-		*tag = encdata + encdatalen - handle->aead.taglen;
-		*taglen = handle->aead.taglen;
-	}
-}
-
 ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
-			   const unsigned char *in, size_t inlen,
+			   unsigned char *in, size_t inlen,
 			   const unsigned char *iv,
-			   const unsigned char *assoc, const unsigned char *tag,
 			   unsigned char *out, size_t outlen, int access)
 {
 	struct iovec iov[3];
 	ssize_t ret = 0;
-	unsigned int bs = handle->info.blocksize;
+	int i = 0;
+	unsigned char *assoc = NULL;
+	size_t assoclen = 0;
+	unsigned char *data = NULL;
+	size_t datalen = 0;
+	unsigned char *tag = NULL;
+	size_t taglen = 0;
 
 	/* require properly sized output data size */
-	if (outlen < _kcapi_aead_decrypt_outlen(handle, inlen)) {
+	if (outlen < inlen) {
 		fprintf(stderr,
-			"AEAD Decryption: Plaintext buffer (%lu) is not ciphertext buffer (%lu) reduced by tag length (%lu) routed up to multiple of block size %u\n",
-			(unsigned long)outlen, (unsigned long) inlen,
-			(unsigned long)handle->aead.taglen, bs);
+			"AEAD Decryption: Plaintext buffer (%lu) is smaller than ciphertext buffer (%lu)\n",
+			(unsigned long)outlen, (unsigned long) inlen);
 		return -EINVAL;
 	}
 
 	handle->cipher.iv = iv;
 
+	kcapi_aead_getdata(handle, in, inlen,
+			   &assoc, &assoclen, &data, &datalen, &tag, &taglen);
+	if (assoclen) {
+		iov[i].iov_base = (void *)assoc;
+		iov[i].iov_len = assoclen;
+		i++;
+	}
+	if (datalen) {
+		iov[i].iov_base = (void *)data;
+		iov[i].iov_len = datalen;
+		i++;
+	}
+	if (taglen) {
+		iov[i].iov_base = (void *)tag;
+		iov[i].iov_len = taglen;
+		i++;
+	}
+
 	if (access == KCAPI_ACCESS_HEURISTIC ||
 	    access == KCAPI_ACCESS_SENDMSG) {
-		if (assoc && handle->aead.assoclen) {
-			iov[0].iov_base = (void*)(uintptr_t)assoc;
-			iov[0].iov_len = handle->aead.assoclen;
-			iov[1].iov_base = (void*)(uintptr_t)in;
-			iov[1].iov_len = inlen;
-			iov[2].iov_base = (void*)(uintptr_t)tag;
-			iov[2].iov_len = handle->aead.taglen;
-			ret = _kcapi_common_send_meta(handle, &iov[0], 3,
-						ALG_OP_DECRYPT, 0);
-		} else {
-			iov[0].iov_base = (void*)(uintptr_t)in;
-			iov[0].iov_len = inlen;
-			iov[1].iov_base = (void*)(uintptr_t)tag;
-			iov[1].iov_len = handle->aead.taglen;
-			ret = _kcapi_common_send_meta(handle, &iov[0], 2,
-						ALG_OP_DECRYPT, 0);
-		}
+		ret = _kcapi_common_send_meta(handle, &iov[0], i,
+					      ALG_OP_DECRYPT, 0);
 		if (0 > ret)
 			return ret;
 	} else {
-		size_t len = 0;
-
 		ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT,
 					MSG_MORE);
 		if (0 > ret)
 			return ret;
-		if (assoc && handle->aead.assoclen) {
-			/* AAD, ciphertext and tag available */
-			iov[0].iov_base = (void*)(uintptr_t)assoc;
-			iov[0].iov_len = handle->aead.assoclen;
-			len = handle->aead.assoclen;
-			iov[1].iov_base = (void*)(uintptr_t)in;
-			iov[1].iov_len = inlen;
-			len += inlen;
-			iov[2].iov_base = (void*)(uintptr_t)tag;
-			iov[2].iov_len = handle->aead.taglen;
-			len += handle->aead.taglen;
-			ret = _kcapi_common_vmsplice_data(handle, &iov[0], 3,
-							  len, 0);
-		} else if (in && inlen) {
-			/* no AAD, but ciphertext and tag available */
-			iov[0].iov_base = (void*)(uintptr_t)in;
-			iov[0].iov_len = inlen;
-			len = inlen;
-			iov[1].iov_base = (void*)(uintptr_t)tag;
-			iov[1].iov_len = handle->aead.taglen;
-			len += handle->aead.taglen;
-			ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2,
-							  len, 0);
-		} else {
-			/* only tag available */
-			iov[0].iov_base = (void*)(uintptr_t)tag;
-			iov[0].iov_len = handle->aead.taglen;
-			len = handle->aead.taglen;
-			ret = _kcapi_common_vmsplice_data(handle, &iov[0], 1,
-							  len, 0);
-		}
+		ret = _kcapi_common_vmsplice_iov(handle, &iov[0], i);
 		if (0 > ret)
 			return ret;
 	}
 
+	/* handle null test vectors */
 	if (outlen) {
 		return _kcapi_common_read_data(handle, out, outlen);
 	} else {
@@ -1031,13 +1069,7 @@ ssize_t kcapi_aead_stream_op(struct kcapi_handle *handle,
 			"AEAD operation: No buffer for output data provided\n");
 		return -EINVAL;
 	}
-#if 0
-	if (iovlen != 1) {
-		fprintf(stderr,
-			"AEAD operation: Output IOV must contain only one entry\n");
-		return -EINVAL;
-	}
-#endif
+
 	return _kcapi_common_recv_data(handle, iov, iovlen);
 }
 
@@ -1057,12 +1089,11 @@ unsigned int kcapi_aead_authsize(struct kcapi_handle *handle)
 }
 
 size_t kcapi_aead_outbuflen(struct kcapi_handle *handle,
-			    size_t inlen, size_t taglen, int enc)
+			    size_t inlen, size_t assoclen, size_t taglen)
 {
-	if (enc)
-		return _kcapi_aead_encrypt_outlen(handle, inlen, taglen);
-	else
-		return _kcapi_aead_decrypt_outlen(handle, inlen);
+	int bs = handle->info.blocksize;
+
+	return ((inlen + bs - 1) / bs * bs + taglen + assoclen);
 }
 
 int kcapi_aead_ccm_nonce_to_iv(const unsigned char *nonce, size_t noncelen,
@@ -1087,14 +1118,15 @@ int kcapi_aead_ccm_nonce_to_iv(const unsigned char *nonce, size_t noncelen,
 	return 0;
 }
 
-int kcapi_md_init(struct kcapi_handle *handle, const char *ciphername)
+int kcapi_md_init(struct kcapi_handle *handle, const char *ciphername,
+		  unsigned int flags)
 {
-	return _kcapi_handle_init(handle, "hash", ciphername);
+	return _kcapi_handle_init(handle, "hash", ciphername, flags);
 }
 
-int kcapi_md_destroy(struct kcapi_handle *handle)
+void kcapi_md_destroy(struct kcapi_handle *handle)
 {
-	return _kcapi_handle_destroy(handle);
+	_kcapi_handle_destroy(handle);
 }
 
 int kcapi_md_setkey(struct kcapi_handle *handle,
@@ -1103,20 +1135,25 @@ int kcapi_md_setkey(struct kcapi_handle *handle,
 	return _kcapi_common_setkey(handle, key, keylen);
 }
 
-static ssize_t _kcapi_md_update(struct kcapi_handle *handle,
-				const unsigned char *buffer, size_t len)
+static inline ssize_t _kcapi_md_update(struct kcapi_handle *handle,
+				       const unsigned char *buffer, size_t len)
 {
 	ssize_t r;
 
-	r = send(handle->opfd, buffer, len, MSG_MORE);
+	/* zero buffer length cannot be handled via splice */
+	/* TODO check that heuristic for sendmsg is appropriate */
+	if(len == 0 /* < (1<<15) */)
+		r = send(handle->opfd, buffer, len, MSG_MORE);
+	else
+		r = _kcapi_common_vmsplice_chunk(handle, buffer, len);
+
 	if (r < 0 || (size_t)r < len)
 		return -EIO;
-
 	return 0;
 }
 
 ssize_t kcapi_md_update(struct kcapi_handle *handle,
-		    const unsigned char *buffer, size_t len)
+			const unsigned char *buffer, size_t len)
 {
 	return _kcapi_md_update(handle, buffer, len);
 }
@@ -1126,7 +1163,7 @@ static ssize_t _kcapi_md_final(struct kcapi_handle *handle,
 {
 	struct iovec iov;
 
-	if (len < (unsigned long)handle->info.hash_digestsize) {
+	if (!buffer || len < (unsigned long)handle->info.hash_digestsize) {
 		fprintf(stderr,
 			"Message digest: output buffer too small (seen %lu - required %u)\n",
 			(unsigned long)len,
@@ -1149,49 +1186,12 @@ ssize_t kcapi_md_digest(struct kcapi_handle *handle,
 		       const unsigned char *in, size_t inlen,
 		       unsigned char *out, size_t outlen)
 {
-	struct iovec iov;
-	size_t processed = 0;
+	ssize_t ret = 0;
 
-	if (!out || !outlen) {
-		fprintf(stderr,
-			"Message digest: Empty plaintext or message digest buffer provided\n");
-		return -EINVAL;
-	}
-
-	if (outlen < (unsigned int)handle->info.hash_digestsize) {
-		fprintf(stderr,
-			"Message digest: output buffer too small (seen %lu - required %u)\n",
-			(unsigned long)outlen,
-			handle->info.hash_digestsize);
-		return -EINVAL;
-	}
-
-	/* zero buffer length cannot be handled via splice */
-	/* TODO check that heuristic for sendmsg is appropriate */
-	if(inlen == 0 /* < (1<<15) */) {
-		if (_kcapi_md_update(handle, in, inlen))
-			return -EIO;
-		return _kcapi_md_final(handle, out, outlen);
-	}
-
-	/* normal zero copy */
-	iov.iov_base = (void*)(uintptr_t)in;
-
-	while (inlen) {
-		ssize_t ret = 0;
-
-		size_t datalen = (inlen > MAXPIPELEN) ? MAXPIPELEN : inlen;
-
-		iov.iov_len = datalen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov, 1, datalen,
-						  SPLICE_F_MORE);
-		if (0 > ret)
-			return ret;
-		processed += ret;
-		iov.iov_base = (void*)(uintptr_t)(in + processed);
-		inlen -= ret;
-	}
-	return _kcapi_common_read_data(handle, out, outlen);
+	ret = _kcapi_md_update(handle, in, inlen);
+	if (0 > ret)
+		return ret;
+	return _kcapi_md_final(handle, out, outlen);
 }
 
 unsigned int kcapi_md_digestsize(struct kcapi_handle *handle)
@@ -1204,14 +1204,15 @@ unsigned int kcapi_md_blocksize(struct kcapi_handle *handle)
 	return handle->info.blocksize;
 }
 
-int kcapi_rng_init(struct kcapi_handle *handle, const char *ciphername)
+int kcapi_rng_init(struct kcapi_handle *handle, const char *ciphername,
+		   unsigned int flags)
 {
-	return _kcapi_handle_init(handle, "rng", ciphername);
+	return _kcapi_handle_init(handle, "rng", ciphername, flags);
 }
 
-int kcapi_rng_destroy(struct kcapi_handle *handle)
+void kcapi_rng_destroy(struct kcapi_handle *handle)
 {
-	return _kcapi_handle_destroy(handle);
+	_kcapi_handle_destroy(handle);
 }
 
 int kcapi_rng_seed(struct kcapi_handle *handle, unsigned char *seed,
