@@ -39,10 +39,28 @@
 
 #include <linux/if_alg.h>
 #include <sys/uio.h>
+#include "kcapi_aio.h"
 
-#define KCAPI_ACCESS_HEURISTIC 0
-#define KCAPI_ACCESS_VMSPLICE  1
-#define KCAPI_ACCESS_SENDMSG   2
+/**
+ * Flags for the encrypt / decrypt operations
+ * 
+ * @KCAPI_ACCESS_HEURISTIC: Allow the libkcapi heuristic to determine the
+ * optimal kernel access type
+ * @KCAPI_ACCESS_VMSPLICE: Require libkcapi to always use the vmsplice zero
+ * copy kernel interface
+ * @KCAPI_ACCESS_SENDMSG: Require libkcapi to always use the sendmsg kernel
+ * interface
+ */
+#define KCAPI_ACCESS_HEURISTIC 	0x0
+#define KCAPI_ACCESS_VMSPLICE  	0x1
+#define KCAPI_ACCESS_SENDMSG   	0x2
+
+/**
+ * Flags for initializing a cipher handle
+ * 
+ * @KCAPI_INIT_AIO: Handle uses AIO kernel interface if available
+ */
+#define KCAPI_INIT_AIO	0x1
 
 /**
  * Information obtained for different ciphers during handle init time
@@ -107,6 +125,7 @@ struct kcapi_aead_data {
  * @skdata: Common data for all ciphers
  * @aead: AEAD cipher specific data
  * @info: properties of ciphers
+ * @aio: AIO information
  */
 struct kcapi_handle {
 	int tfmfd;
@@ -115,6 +134,7 @@ struct kcapi_handle {
 	struct kcapi_cipher_data cipher;
 	struct kcapi_aead_data aead;
 	struct kcapi_cipher_info info;
+	struct kcapi_aio aio;
 };
 
 /**
@@ -128,6 +148,7 @@ struct kcapi_handle {
  * @handle: cipher handle filled during the call - output
  * @ciphername: kernel crypto API cipher name as specified in
  *	       /proc/crypto - input
+ * @flags: flags specifying the type of cipher handle
  *
  * This function provides the initialization of a symmetric cipher handle and
  * establishes the connection to the kernel.
@@ -136,15 +157,14 @@ struct kcapi_handle {
  *	   -EOPNOTSUPP - AF_ALG family not available;
  *	   -EINVAL - accept syscall failed
  */
-int kcapi_cipher_init(struct kcapi_handle *handle, const char *ciphername);
+int kcapi_cipher_init(struct kcapi_handle *handle, const char *ciphername,
+		      unsigned int flags);
 
 /**
  * kcapi_cipher_destroy() - close the cipher handle and release resources
  * @handle: cipher handle to release - input
- *
- * Return: 0 upon success
  */
-int kcapi_cipher_destroy(struct kcapi_handle *handle);
+void kcapi_cipher_destroy(struct kcapi_handle *handle);
 
 /**
  * kcapi_cipher_setkey() - set the key for the cipher handle
@@ -367,6 +387,7 @@ unsigned int kcapi_cipher_blocksize(struct kcapi_handle *handle);
  * @handle: cipher handle filled during the call - output
  * @ciphername: kernel crypto API cipher name as specified in
  *	       /proc/crypto - input
+ * @flags: flags specifying the type of cipher handle
  *
  * This function initializes an AEAD cipher handle and establishes the
  * connection to the kernel.
@@ -376,15 +397,14 @@ unsigned int kcapi_cipher_blocksize(struct kcapi_handle *handle);
  *	   -EOPNOTSUPP - AF_ALG family not available;
  *	   -EINVAL - accept syscall failed
  */
-int kcapi_aead_init(struct kcapi_handle *handle, const char *ciphername);
+int kcapi_aead_init(struct kcapi_handle *handle, const char *ciphername,
+		    unsigned int flags);
 
 /**
  * kcapi_aead_destroy() - close the AEAD handle and release resources
  * @handle: cipher handle to release - input
- *
- * Return: 0 upon success
  */
-int kcapi_aead_destroy(struct kcapi_handle *handle);
+void kcapi_aead_destroy(struct kcapi_handle *handle);
 
 /**
  * kcapi_aead_setkey() - set the key for the AEAD handle
@@ -436,7 +456,6 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, size_t assoclen);
  * @in: plaintext data buffer - input
  * @inlen: length of plaintext buffer - input
  * @iv: IV to be used for cipher operation - input
- * @assoc: associated data of size set with kcapi_aead_setassoclen() - input
  * @out: data buffer holding cipher text and authentication tag - output
  * @outlen: length of out buffer - input
  * @access: kernel access type (KCAPI_ACCESS_HEURISTIC - use internal heuristic
@@ -465,16 +484,20 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, size_t assoclen);
  *	   < 0 in case of error with errno set
  */
 ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
-			   const unsigned char *in, size_t inlen,
+			   unsigned char *in, size_t inlen,
 			   const unsigned char *iv,
-			   const unsigned char *assoc, unsigned char *out,
-			   size_t outlen, int access);
+			   unsigned char *out, size_t outlen,
+			   int access);
 
 /**
  * kcapi_aead_getdata() - Get the resulting data from encryption
  * @handle: cipher handle - input
  * @encdata: data buffer returned by the encryption operation - input
  * @encdatalen: size of the encryption data buffer - input
+ * @aad: AD buffer pointer;  when set to NULL, no data pointer is returned
+ *	 - output
+ * @aadlen: length of AD; when @aad was set to NULL, no information is returned
+ *	    - output
  * @data: pointer to output buffer from AEAD encryption operation when set to
  *	  NULL, no data pointer is returned - output
  * @datalen: length of data buffer; when @data was set to NULL, no information
@@ -491,6 +514,7 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
  */
 void kcapi_aead_getdata(struct kcapi_handle *handle,
 			unsigned char *encdata, size_t encdatalen,
+			unsigned char **aad, size_t *aadlen,
 			unsigned char **data, size_t *datalen,
 			unsigned char **tag, size_t *taglen);
 
@@ -500,8 +524,6 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
  * @in: ciphertext data buffer - input
  * @inlen: length of in buffer - input
  * @iv: IV to be used for cipher operation - input
- * @assoc: associated data of size set with kcapi_aead_setassoclen() - input
- * @tag: authentication tag data of size set with kcapi_aead_settaglen() - input
  * @out: plaintext data buffer - output
  * @outlen: length of out buffer - input
  * @access: kernel access type (KCAPI_ACCESS_HEURISTIC - use internal heuristic
@@ -531,9 +553,8 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
  *	   < 0 in case of error with errno set
  */
 ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
-			   const unsigned char *in, size_t inlen,
+			   unsigned char *in, size_t inlen,
 			   const unsigned char *iv,
-			   const unsigned char *assoc, const unsigned char *tag,
 			   unsigned char *out, size_t outlen, int access);
 
 /**
@@ -723,15 +744,14 @@ unsigned int kcapi_aead_authsize(struct kcapi_handle *handle);
 /**
  * kcapi_aead_outbuflen() - return minimum output buffer length
  * @handle: cipher handle - input
- * @inlen: size of plaintext (if @enc is one) or size of ciphertext (if @enc
- * 	   is zero)
+ * @inlen: size of plaintext or size of ciphertext
+ * @assoclen: size of associated data (AD)
  * @taglen: size of authentication tag
- * @enc: type of cipher operation (1 == encryption, 0 == decryption)
  *
  * Return: minimum size of output data length in bytes
  */
 size_t kcapi_aead_outbuflen(struct kcapi_handle *handle,
-			    size_t inlen, size_t taglen, int enc);
+			    size_t inlen, size_t assoclen, size_t taglen);
 
 /**
  * kcapi_aead_ccm_nonce_to_iv() - convert CCM nonce into IV
@@ -761,6 +781,7 @@ int kcapi_aead_ccm_nonce_to_iv(const unsigned char *nonce, size_t noncelen,
  * @handle: cipher handle filled during the call - output
  * @ciphername: kernel crypto API cipher name as specified in
  *	       /proc/crypto - input
+ * @flags: flags specifying the type of cipher handle
  *
  * This function provides the initialization of a (keyed) message digest handle
  * and establishes the connection to the kernel.
@@ -769,15 +790,14 @@ int kcapi_aead_ccm_nonce_to_iv(const unsigned char *nonce, size_t noncelen,
  *	   -EOPNOTSUPP - AF_ALG family not available;
  *	   -EINVAL - accept syscall failed
  */
-int kcapi_md_init(struct kcapi_handle *handle, const char *ciphername);
+int kcapi_md_init(struct kcapi_handle *handle, const char *ciphername,
+		  unsigned int flags);
 
 /**
  * kcapi_md_destroy() - close the message digest handle and release resources
  * @handle: cipher handle to release - input
- *
- * Return: 0 upon success
  */
-int kcapi_md_destroy(struct kcapi_handle *handle);
+void kcapi_md_destroy(struct kcapi_handle *handle);
 
 /**
  * kcapi_md_setkey() - set the key for the message digest handle
@@ -879,6 +899,7 @@ unsigned int kcapi_md_blocksize(struct kcapi_handle *handle);
  * @handle: cipher handle filled during the call - output
  * @ciphername: kernel crypto API cipher name as specified in
  *	       /proc/crypto - input
+ * @flags: flags specifying the type of cipher handle
  *
  * This function provides the initialization of a random number generator handle
  * and establishes the connection to the kernel.
@@ -887,15 +908,14 @@ unsigned int kcapi_md_blocksize(struct kcapi_handle *handle);
  *	   -EOPNOTSUPP - AF_ALG family not available;
  *	   -EINVAL - accept syscall failed
  */
-int kcapi_rng_init(struct kcapi_handle *handle, const char *ciphername);
+int kcapi_rng_init(struct kcapi_handle *handle, const char *ciphername,
+		   unsigned int flags);
 
 /**
  * kcapi_rng_destroy() - Close the RNG handle and release resources
  * @handle: cipher handle to release - input
- *
- * Return: 0 upon success
  */
-int kcapi_rng_destroy(struct kcapi_handle *handle);
+void kcapi_rng_destroy(struct kcapi_handle *handle);
 
 /**
  * kcapi_rng_seed() - Seed the RNG
@@ -971,4 +991,12 @@ int kcapi_pad_iv(struct kcapi_handle *handle,
 		 const unsigned char *iv, size_t ivlen,
 		 unsigned char **newiv, size_t *newivlen);
 
+/**
+ * kcapi_memset_secure() - memset() implementation that will not be optimized
+ *			   away by the compiler
+ * 
+ * The parameters, he logic and the return code is identical to memset(3).
+ */
+void kcapi_memset_secure(void *s, int c, size_t n);
+
 #endif /* _KCAPI_H */
